#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ctype.h>
#include <gtk/gtk.h>
#include <svn_client.h>

#include "subview.h"

enum logcol_t {
	LOGCOL_REV,
	LOGCOL_AUTHOR,
	LOGCOL_MESSAGE,
	LOGCOL_DATE,
	LOGCOL_CHANGED_PATHS,
	LOGNCOLS
};

enum {
	CP_ST,
	CP_PATH,
	CPNCOLS
};

struct log_window {
	GtkWidget *window;
	GtkListStore *model;
	apr_pool_t *pool;
};

svn_error_t* log_receiver(void *baton, apr_hash_t *changed_paths, svn_revnum_t revision, const char *author, const char *date,const char *message, apr_pool_t *pool)
{
	struct log_window *y = baton;
	GtkTreeIter it;
	const char *p, *next_p;
	char m[4096], *q;
	int n, i = 0;
	
	q = m;
	
	p=message;
	while(p && *p && i<4)
	{
		if(q!=m)
			*q++ = '\n';
		n = strcspn(p, "\n");
		next_p = p + n + 1;
		while(n && isspace(*(p+n-1)))
			n--;
		strncpy(q,p,n);
		q+=n;
		if(!*(p+n))
			break;
		p=next_p;
		if(*p == '\r')
			p++;
		i++;
	}
	*q = '\0';
	
	//d = get_date_string();

	gdk_threads_enter();
	
	gtk_list_store_append(y->model, &it);
	gtk_list_store_set(y->model, &it
		, LOGCOL_REV, (long)revision
		, LOGCOL_AUTHOR, author
		, LOGCOL_MESSAGE, m
		, LOGCOL_DATE, date
		, -1);

	if(changed_paths)
	{
		apr_hash_index_t *hi;
		GtkListStore *changed_paths_list = gtk_list_store_new(CPNCOLS, G_TYPE_STRING, G_TYPE_STRING);
		GtkTreeIter li;

		for(hi = apr_hash_first(pool, changed_paths) ; hi ; hi = apr_hash_next(hi))
		{
			const char *key;
			char action[2];
			svn_log_changed_path_t *val;
			apr_hash_this(hi, (void*)&key, NULL, (void*)&val);
			gtk_list_store_append(changed_paths_list, &li);
			action[0] = val->action;
			action[1] = '\0';
			gtk_list_store_set(changed_paths_list, &li
				, CP_ST, action
				, CP_PATH, key, -1);
		}
		gtk_list_store_set(y->model, &it
			, LOGCOL_CHANGED_PATHS, changed_paths_list, -1);
	}
	gdk_threads_leave();
	//g_free(d);
	return NULL;
}

void get_log(work_t *work)
{
	svn_error_t *err;
	svn_opt_revision_t start =
	{
		.kind = svn_opt_revision_number,
		.value = { .number = 1 }
	};

	apr_array_header_t *targets;
	targets = apr_array_make(work->y->pool, 1, sizeof(char*));
	
	((const char**)(targets->elts))[0] = work->url;
	targets->nelts = 1;

	err = svn_client_log2(targets, &HEAD, &start, 0, TRUE, FALSE,
		&log_receiver, work->y, ctx, work->y->pool);
	if(err)
		svn_handle_error(err, stderr, FALSE);

	return;
}

static void on_cursor_changed(GtkTreeView *tv_revs, gpointer data)
{
	GtkTreeModel *model, *cpmodel;
	GtkTreeIter i;
	GtkTreeView *tv_paths = (GtkTreeView *)data;

	if(!gtk_tree_selection_get_selected(gtk_tree_view_get_selection(tv_revs), &model, &i))
		return;
	gtk_tree_model_get(model, &i, LOGCOL_CHANGED_PATHS, &cpmodel, -1);
	gtk_tree_view_set_model(tv_paths, cpmodel);
}

/** Shows a file's log.
 */
void show_log(GtkMenuItem *menuitem, gpointer data)
{
	struct log_window *y;
	apr_pool_t *pool;
	char msg[2048];
	GtkTreeViewColumn *col;
	GtkCellRenderer *r;
	GtkWidget *dialog, *tv_revs, *tv_paths, *sw, *vp;

	apr_pool_create(&pool, NULL);

	y = g_new(struct log_window, 1);

	work_t *work = g_new(work_t, 1);
	work->url = (char*)g_object_get_data(G_OBJECT(menuitem), "repos_url");
	g_return_if_fail(work->url);

	sprintf(msg, _("Log of '%s'"), work->url);
	dialog = gtk_dialog_new_with_buttons(msg, GTK_WINDOW(data), GTK_DIALOG_DESTROY_WITH_PARENT, GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);

	g_signal_connect_swapped (dialog, "response", G_CALLBACK(gtk_widget_destroy), dialog);

	//gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), gtk_label_new("Log"), FALSE, FALSE, 0);

	y->model = gtk_list_store_new(LOGNCOLS, G_TYPE_LONG, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, GTK_TYPE_LIST_STORE);
	tv_revs = gtk_tree_view_new_with_model(GTK_TREE_MODEL(y->model));

	r = gtk_cell_renderer_text_new();
	/*
	memset(&gtrue, 0, sizeof(gtrue));
	g_value_init(&gtrue, G_TYPE_BOOLEAN);
	g_value_set_boolean(&gtrue, TRUE);
	g_object_set_property(G_OBJECT(r), "ellipsize", &gtrue);
	g_value_unset(&gtrue);
	*/
	col = gtk_tree_view_column_new_with_attributes(_("Rev"), r, "text", LOGCOL_REV, NULL);
	gtk_tree_view_column_set_sort_column_id(col, LOGCOL_REV);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv_revs), col);
	col = gtk_tree_view_column_new_with_attributes(_("Author"), r, "text", LOGCOL_AUTHOR, NULL);
	gtk_tree_view_column_set_sort_column_id(col, LOGCOL_AUTHOR);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv_revs), col);
	col = gtk_tree_view_column_new_with_attributes(_("Date"), r, "text", LOGCOL_DATE, NULL);
	gtk_tree_view_column_set_sort_column_id(col, LOGCOL_DATE);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv_revs), col);
	col = gtk_tree_view_column_new_with_attributes(_("Message"), r, "text", LOGCOL_MESSAGE, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv_revs), col);	

	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(sw), tv_revs);
	
	vp = gtk_vpaned_new();
	gtk_paned_pack1(GTK_PANED(vp), sw, TRUE, FALSE);

	tv_paths = gtk_tree_view_new();
	col = gtk_tree_view_column_new_with_attributes(_("Action"), r, "text", CP_ST, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv_paths), col);
	col = gtk_tree_view_column_new_with_attributes(_("Path"), r, "text", CP_PATH, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv_paths), col);
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(sw), tv_paths);
	gtk_paned_pack2(GTK_PANED(vp), sw, TRUE, FALSE);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), vp, TRUE, TRUE, 0);

	g_signal_connect(G_OBJECT(tv_revs), "cursor-changed", G_CALLBACK(on_cursor_changed), tv_paths);
	
	gtk_window_set_default_size(GTK_WINDOW(dialog), 500, 400);

	gtk_widget_show_all(dialog);
	
	work->operation = op_log;
	work->y = y;
	y->pool = pool;
	work->window = y->window;
	
	g_thread_pool_push(thread_pool, work, NULL);
}
