#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <locale.h>
#include <libintl.h>

#include <gtk/gtk.h>
#include <svn_client.h>

#include "subview.h"

svn_client_ctx_t *ctx;

GThreadPool *thread_pool;

GStaticPrivate thread_main_window = G_STATIC_PRIVATE_INIT;

struct browse_window {
	GtkWidget *window, *entry;
	GtkTreeStore *model;
	apr_pool_t *pool;
	int model_number;
};

typedef enum {
	COL_URL,
	COL_NAME,
	COL_LASTREV,
	COL_LASTAUTHOR,
	COL_LASTMOD,
	COL_CHILDREN_LOADED,
	NCOLS
} col_t;

void expand(GtkTreeView *treeview, GtkTreeIter *i, GtkTreePath *p, gpointer data)
{
	char *url;
	gboolean children_loaded;
	struct browse_window *x = (struct browse_window *)data;
	work_t *work = g_new(work_t, 1);

	work->operation = op_log;
	gtk_tree_model_get(GTK_TREE_MODEL(x->model), i
		, COL_URL, &url
		, COL_CHILDREN_LOADED, &children_loaded
		, -1);
	if(children_loaded)
		return;
	gtk_tree_store_set(x->model, i
		,COL_CHILDREN_LOADED, TRUE
		,-1);
	work->url = url;
	work->operation = op_get_dir;
	work->model_number = x->model_number;
	work->x = x;
	work->where = *i;
	work->window = x->window;
	g_thread_pool_push(thread_pool, work, NULL);
}

static void browse(GtkButton *button, gpointer data)
{
	GtkTreeIter i, p;
	const char *url;
	int pos, n = 0;
	char name[1024];
	struct browse_window *x = (struct browse_window *)data;
	work_t *work = g_new(work_t, 1);

	url = gtk_entry_get_text(GTK_ENTRY(x->entry));
	gtk_tree_store_clear(x->model);
	x->model_number++;
	pos = strlen(url) - 1;
	if(url[pos] == '/')
	{
		strncpy(name, url, pos);
		name[pos] = '\0';
		gtk_entry_set_text(GTK_ENTRY(x->entry), name);
		url = gtk_entry_get_text(GTK_ENTRY(x->entry));
		pos--;
	}
	while(url[pos] != '/')
	{
		n++;
		pos--;
	}
	strncpy(name, url + pos + 1, n);
	name[n] = '\0';

	gtk_tree_store_insert(x->model, &i, NULL, 0);
	gtk_tree_store_set(x->model, &i
		,COL_URL, url
		,COL_NAME, name
		,COL_CHILDREN_LOADED, FALSE
		,-1);

	work->where = i;
	work->url = g_strdup(url);
	work->x = x;
	work->operation = op_get_root;
	work->model_number = x->model_number;
	work->window = x->window;
	p = i;
	gtk_tree_store_insert(x->model, &i, &p, 0);
	gtk_tree_store_set(work->x->model, &i
		,COL_NAME, _("<loading>")
		,-1);

	g_thread_pool_push(thread_pool, work, NULL);
}

void cleanup_repobrowser(GtkObject *object, gpointer data)
{
	struct browse_window *x = (struct browse_window *)data;
	g_free(x);
}

gboolean context_menu(GtkWidget *widget, GdkEventButton *event, gpointer data)
{
	struct browse_window *x = (struct browse_window *)data;
	char msg[1024], *url, *filename;
	GtkWidget *menu, *i;
	GtkTreePath *path;
	gint cx, cy;
	GtkTreeIter it;

	if(event->button != 3)
		return FALSE;
	
	if(!gtk_tree_view_get_path_at_pos(GTK_TREE_VIEW(widget), event->x, event->y, &path, NULL, &cx, &cy))
		return FALSE;

	gtk_tree_selection_select_path(gtk_tree_view_get_selection(GTK_TREE_VIEW(widget)), path);
	gtk_tree_model_get_iter(GTK_TREE_MODEL(x->model), &it, path);
	gtk_tree_path_free(path);
	gtk_tree_model_get(GTK_TREE_MODEL(x->model), &it
		, COL_URL, &url
		, COL_NAME, &filename
		, -1);
	
	sprintf(msg, _("Show log for '%s'..."), filename);
	g_free(filename);
	
	menu = gtk_menu_new();

	i = gtk_menu_item_new_with_label(msg);
	g_object_set_data_full(G_OBJECT(i), "repos_url", url, g_free);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), i);
	gtk_widget_show(i);
	g_signal_connect(G_OBJECT(i), "activate", G_CALLBACK(show_log), x->window);

	i = gtk_menu_item_new_with_label(_("Checkout..."));
	g_object_set_data_full(G_OBJECT(i), "repos_url", url, g_free);
	gtk_menu_shell_append(GTK_MENU_SHELL(menu), i);
	gtk_widget_show(i);
	g_signal_connect(G_OBJECT(i), "activate", G_CALLBACK(checkout), x->window);

	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, NULL, NULL, event->button, event->time);
	
	return TRUE;
}

GtkWidget *create_repo_browser()
{
	struct browse_window *x;
	GtkWidget *vb, *hb, *tv, *sw, *b;
	GtkTreeViewColumn *col;
	GtkCellRenderer *r;

	x = g_new(struct browse_window, 1);
	x->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	x->model_number = 0;
	apr_pool_create(&(x->pool), NULL);
	
	gtk_window_set_title(GTK_WINDOW(x->window), _("Repository browser"));
	
	vb = gtk_vbox_new(FALSE, 3);
	
	gtk_container_add(GTK_CONTAINER(x->window), vb);
	
	hb = gtk_hbox_new(FALSE, 3);
	gtk_box_pack_start(GTK_BOX(hb), gtk_label_new(_("URL:")), FALSE, TRUE, 0);
	x->entry = gtk_entry_new();
	gtk_entry_set_text(GTK_ENTRY(x->entry), "http://www.reloco.com.ar/svn/barras/tronco");
	gtk_box_pack_start(GTK_BOX(hb), x->entry, TRUE, TRUE, 0);
	b = gtk_button_new_with_label(_("Browse"));
	gtk_box_pack_start(GTK_BOX(hb), b, FALSE, TRUE, 0);

	gtk_box_pack_start(GTK_BOX(vb), hb, FALSE, TRUE, 0);
	
	x->model = gtk_tree_store_new(NCOLS, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);
	tv = gtk_tree_view_new_with_model(GTK_TREE_MODEL(x->model));
	gtk_tree_view_enable_model_drag_source(GTK_TREE_VIEW(tv), 0, NULL, 0, GDK_ACTION_LINK|GDK_ACTION_DEFAULT);
	r = gtk_cell_renderer_text_new();
	col = gtk_tree_view_column_new_with_attributes(_("Filename"), r, "text", COL_NAME, NULL);
	gtk_tree_view_set_search_column(GTK_TREE_VIEW(tv), COL_NAME);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv), col);
	col = gtk_tree_view_column_new_with_attributes(_("Last modified"), r, "text", COL_LASTMOD, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv), col);
	col = gtk_tree_view_column_new_with_attributes(_("Modified in rev"), r, "text", COL_LASTREV, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv), col);
	col = gtk_tree_view_column_new_with_attributes(_("Last author"), r, "text", COL_LASTAUTHOR, NULL);
	gtk_tree_view_append_column (GTK_TREE_VIEW(tv), col);
	gtk_tree_view_set_enable_search(GTK_TREE_VIEW(tv), TRUE);
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(sw), tv);
	gtk_box_pack_start(GTK_BOX(vb), sw, TRUE, TRUE, 0);

	g_signal_connect(G_OBJECT(b), "clicked", G_CALLBACK(browse), x);
	g_signal_connect(G_OBJECT(tv), "row-expanded", G_CALLBACK(expand), x);
	g_signal_connect(G_OBJECT(tv), "destroy", G_CALLBACK(cleanup_repobrowser), x);
	g_signal_connect(G_OBJECT(tv), "button_press_event", G_CALLBACK(context_menu), x);
	g_signal_connect(G_OBJECT(x->entry), "activate", G_CALLBACK(browse), x);
	
	gtk_window_set_default_size(GTK_WINDOW(x->window), 500, 300);

	return x->window;
}

svn_error_t *info_receiver(void *baton, const char *path, const svn_info_t *info, apr_pool_t *pool)
{
	work_t *work = (work_t*)baton;

	if(work->x->model_number == work->model_number)
	{
		char z[40];
		gdk_threads_enter();
		sprintf(z, "%ld", info->last_changed_rev);
		gtk_tree_store_set(work->x->model, &(work->where)
			,COL_LASTAUTHOR, info->last_changed_author
			,COL_LASTREV, z
			,-1);
		gdk_threads_leave();
	}

	return NULL;
}

svn_opt_revision_t HEAD =
{
	.kind = svn_opt_revision_head
};

void get_root(work_t *work)
{
	svn_error_t *err = svn_client_info(work->url, &HEAD, &HEAD
		, info_receiver, work
		, FALSE, ctx, work->x->pool);
	if(err)
	{
		svn_handle_error2(err, stderr, FALSE, "svn: ");
		
	}
}

char *get_date_string(apr_time_t t)
{
	apr_time_exp_t tm;
	char lastmod[100];
	apr_size_t tam = sizeof(lastmod);
	apr_status_t st;

	apr_time_exp_lt(&tm, t);
	st = apr_strftime(lastmod, &tam, sizeof(lastmod), "%c", &tm);
	if(st)
	{
		apr_strerror(st, lastmod, sizeof(lastmod));
		g_warning(lastmod);
	}
	return g_locale_to_utf8(lastmod, -1, NULL, NULL, NULL);
}

void get_dir(work_t *work)
{
	apr_hash_t *dirents; //, *locks;

	svn_error_t *err = svn_client_ls3(&dirents, NULL /* &locks */, work->url, &HEAD, &HEAD, FALSE, ctx, work->x->pool);
	
	if(err)
	{
		svn_handle_error2(err, stderr, FALSE, "svn: ");
		gtk_tree_store_set(work->x->model, &(work->where)
			,COL_CHILDREN_LOADED, FALSE
			,-1);
		return;
	}
	if(work->x->model_number != work->model_number)
		return;

	GtkTreeIter it;
	apr_hash_index_t *i = apr_hash_first(work->x->pool, dirents);

	gdk_threads_enter();
	
	gtk_tree_model_iter_children(GTK_TREE_MODEL(work->x->model), &it, &(work->where));
	
	while(i)
	{
		const char *k;
		const svn_dirent_t *d;
		char z[40];
		char *str, *str_url;

		apr_hash_this(i, (void*)&k, NULL, (void*)&d);

		str = get_date_string(d->time);
		
		sprintf(z, "%ld", d->created_rev);
		str_url = g_strconcat(work->url, "/", k, NULL);
		gtk_tree_store_set(work->x->model, &it
			,COL_URL, str_url
			,COL_NAME, k
			,COL_LASTAUTHOR, d->last_author
			,COL_LASTREV, z
			,COL_LASTMOD, str
			,-1);
		g_free(str);
		g_free(str_url);
		
		if(d->kind == svn_node_dir)
		{
			GtkTreeIter it2;
			gtk_tree_store_insert(work->x->model, &it2, &it, 0);
			gtk_tree_store_set(work->x->model, &it2
				,COL_NAME, _("<loading>")
				,-1);
		}
		
		i = apr_hash_next(i);
		if(i)
			gtk_tree_store_append(work->x->model, &it, &(work->where));
	}
	gdk_threads_leave();
}

void worker_thread(void *data, void *zz)
{
	work_t *work = (work_t*)data;
	
	g_static_private_set(&thread_main_window, work->window, NULL);

	switch(work->operation)
	{
		case op_get_root:
			get_root(work);
			break;
		case op_get_dir:
			get_dir(work);
			break;
		case op_log:
			get_log(work);
			break;
		default:
			g_error("unknown op %d!", work->operation);
			break;
	}

	g_static_private_set(&thread_main_window, NULL, NULL);

	g_free(work->url);
	g_free(work);
}

static svn_error_t *
ask_password(svn_auth_cred_simple_t **cred,
                           void *baton,
                           const char *realm,
                           const char *username,
                           svn_boolean_t may_save,
                           apr_pool_t *pool)
{
	GtkWidget *dialog, *hb, *vb, *l, *user_entry = NULL, *password_entry, *label;
	GtkWidget *window = g_static_private_get(&thread_main_window);
	GtkSizeGroup *sg;
	char *msg;
	int res;
	svn_auth_cred_simple_t *ret = apr_pcalloc (pool, sizeof (*ret));

	/* printf("r=%s, u=%s, may_save=%d\n", realm, username, may_save); */
	gdk_threads_enter();
	dialog = gtk_dialog_new_with_buttons(_("Logging in")
		, GTK_WINDOW(window), GTK_DIALOG_MODAL
		, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL
		, GTK_STOCK_OK, GTK_RESPONSE_OK
		, NULL);
	gtk_dialog_set_default_response(GTK_DIALOG(dialog), GTK_RESPONSE_OK);
	gtk_dialog_set_has_separator(GTK_DIALOG(dialog), FALSE);
	gtk_window_set_resizable(GTK_WINDOW(dialog), FALSE);
	gtk_container_set_border_width(GTK_CONTAINER(dialog), 6);

	hb = gtk_hbox_new(FALSE, 12);
	gtk_container_set_border_width(GTK_CONTAINER(hb), 6);

	label = gtk_image_new_from_stock(GTK_STOCK_DIALOG_AUTHENTICATION, GTK_ICON_SIZE_DIALOG);
	gtk_misc_set_alignment(GTK_MISC(label), .5, 0);
	gtk_box_pack_start(GTK_BOX(hb), label, FALSE, TRUE, 0);

	vb = gtk_vbox_new(FALSE, 3);
	
	sg = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
	
	label = gtk_label_new(_("<span weight=\"bold\" size=\"larger\">Password required</span>"));
	gtk_label_set_use_markup(GTK_LABEL(label), TRUE);
	gtk_misc_set_alignment(GTK_MISC(label), 0, .5);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_box_pack_start(GTK_BOX(vb), label, TRUE, TRUE, 6);

	msg = g_strdup_printf(_("Enter your password for the Subversion repository at %s."), realm);
	label = gtk_label_new(msg);
	gtk_misc_set_alignment(GTK_MISC(label), 0, .5);
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE);
	gtk_box_pack_start(GTK_BOX(vb), label, TRUE, TRUE, 0);
	
	if(!username)
	{
		l = gtk_hbox_new(FALSE, 3);
		label = gtk_label_new_with_mnemonic(_("_User:"));
		gtk_misc_set_alignment(GTK_MISC(label), 0, .5);
		gtk_box_pack_start(GTK_BOX(l), label, FALSE, FALSE, 0);
		user_entry = gtk_entry_new();
		gtk_label_set_mnemonic_widget(GTK_LABEL(label), user_entry);
		gtk_size_group_add_widget(sg, label);
		gtk_box_pack_start(GTK_BOX(l), user_entry, TRUE, TRUE, 0);
		gtk_box_pack_start(GTK_BOX(vb), l, TRUE, TRUE, 0);
	}

	l = gtk_hbox_new(FALSE, 3);
	label = gtk_label_new_with_mnemonic(_("_Password:"));
	gtk_misc_set_alignment(GTK_MISC(label), 0, .5);
	gtk_box_pack_start(GTK_BOX(l), label, FALSE, FALSE, 0);
	password_entry = gtk_entry_new();
	gtk_label_set_mnemonic_widget(GTK_LABEL(label), password_entry);
	gtk_entry_set_activates_default(GTK_ENTRY(password_entry), TRUE);
	gtk_size_group_add_widget(sg, label);
	gtk_entry_set_visibility(GTK_ENTRY(password_entry), FALSE);
	gtk_box_pack_start(GTK_BOX(l), password_entry, TRUE, TRUE, 0);
	gtk_box_pack_start(GTK_BOX(vb), l, TRUE, TRUE, 0);

	g_object_unref(G_OBJECT(sg));

	gtk_box_pack_start(GTK_BOX(hb), vb, TRUE, TRUE, 0);

	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(dialog)->vbox), hb, TRUE, TRUE, 0);
	
	gtk_widget_show_all(dialog);

	res = gtk_dialog_run(GTK_DIALOG(dialog));
	
	if(res != GTK_RESPONSE_OK)
	{
		gtk_widget_destroy(dialog);
		gdk_threads_leave();
		return svn_error_create(SVN_ERR_AUTHN_CREDS_UNAVAILABLE, NULL, NULL);
	}

	if(username)
	{
		ret->username = apr_pstrdup(pool, username);
	} else
	{
		ret->username = apr_pstrdup(pool, gtk_entry_get_text(GTK_ENTRY(user_entry)));
	}
		
	ret->password = apr_pstrdup(pool, gtk_entry_get_text(GTK_ENTRY(password_entry)));

	gtk_widget_destroy(dialog);

	gdk_threads_leave();

	*cred = ret;
	return NULL;
}


int main(int argc, char **argv)
{
	svn_error_t *err;
	apr_pool_t *pool;
	GtkWidget *w;
	svn_auth_provider_object_t *ap;
	apr_array_header_t *providers;
	
	bindtextdomain(PACKAGE, LOCALEDIR);
	bind_textdomain_codeset(PACKAGE, "UTF-8");
	textdomain(PACKAGE);
	
	g_thread_init(NULL);
	gdk_threads_init();
	gdk_threads_enter();
	
	apr_app_initialize(&argc, (const char* const**)&argv, NULL);
	
	apr_pool_create(&pool, NULL);
	err = svn_client_create_context(&ctx, pool);
	if(err)
		svn_handle_error(err, stderr, FALSE);

	providers = apr_array_make(pool, 2, sizeof(svn_auth_provider_object_t*));
	
	/*
	svn_client_get_simple_provider(&ap, pool);
	((svn_auth_provider_object_t**)(providers->elts))[0] = ap;
	
	my_auth_provider.provider_baton = NULL;
	my_auth_provider.vtable = NULL;	
	((svn_auth_provider_object_t**)(providers->elts))[1] = &my_auth_provider;
	providers->nelts = 2;
	*/
	
	svn_auth_get_simple_prompt_provider(&ap, &ask_password, NULL, 20, pool);

	((svn_auth_provider_object_t**)(providers->elts))[0] = ap;

	providers->nelts = 1;

	svn_auth_open(&(ctx->auth_baton), providers, pool);
	gtk_init(&argc, &argv);

	/*	
	queue = g_async_queue_new();
	if(!g_thread_create(worker_thread, NULL, FALSE, NULL))
	{
		return 1;
	}*/
	
	thread_pool = g_thread_pool_new(worker_thread, NULL, 4, FALSE, NULL);
	g_thread_pool_set_max_unused_threads(-1);
	
	w = create_repo_browser();
	
	gtk_widget_show_all(w);
	
	g_signal_connect(G_OBJECT(w), "destroy", G_CALLBACK(gtk_main_quit), NULL);

/*
	err = dumplog(pool, argc>1 ? argv[1] : "file:///home/nick/src/subview/repos/cosas/tronco");
	if(err)
		svn_handle_error(err, stderr, FALSE);
*/

	gtk_main();
	
	gdk_threads_leave();

	apr_terminate();

	return 0;
}
